function [dico,freq,repcand,repfreq,Sguess,avenres,avsc2]=itkrmplus_1it(dico,data,S,repinit,valtype,minobs)

% syntax: [dico, ...] = itkrmplus_1it(dico,data,S,repinit,valtype,minobs)
%
% one iteration of the Iterative Thresholding& 
% K residual Means dictionary learning algorithm 
% modified for replacement and adaptive dictionary learning
% as described in arXiv:1804.07101
% 
% input:
% dico... d x K matrix containing the current dictionary estimate
% data... d x N matrix containing the training signals as its columns
% S... sparsity level used for thresholding - default 1
% repinit... initialisation of candidate atoms - default random
% minobs... minimal number required observations - default: d*log(d)
% valtype... type of value function - default: counting above threshold
%
% output:
% dico... d x K dictionary 
% freq... score for each atom in the dictionary
% repcand... candidate atoms
% repfreq... score of candidate atoms
% Sguess... [estimated sparsity level, av. of correctly identified atoms]
% avenres... average energy of the residual
% avsc2... average coefficients/inner products sorted and squared
%
% Karin Schnass 28.05.18 

%%%% preparations and basic safeguarding
if(nargin < 2)
    disp('[dico,...]=itkrm_1it(dico,data,S,repinit,minobs,valtype)');
    dico=[];
    return;
end

[d,N] = size(data);
[dl,K] = size(dico);

if (dl ~= d)
    disp('dimensions of signals and dictionary do not match');
    return;
end

if (N < K+1)
    disp('less training signals than atoms => trivial solution');
    dico=data;
    return;
end

if(nargin < 3)
    S=1;
end

if(nargin < 4) 
    L = round(log(d));
    repinit = randn(d,L); 
    scale = sum(repinit.*repinit);
    repinit = repinit*diag(1./sqrt(scale));
else
    L=size(repinit,2);
end

if(nargin < 5)
    valtype = 0;
    minobs = round(d*log(d));
end

if ~ismember(valtype,[-1,0,2])
    valtype = 0;
    minobs = round(d*log(d));
end

if(nargin < 6) && valtype >= 0 
    minobs = round(d*log(d));
end

m = round(log(d));   %%% number of candidate iterations/dico it.
NL=floor(N/m);  

if valtype == -1
    candtau = 2*log(2*K)/d;  %%% simple threshold for replacement
else
    candtau = 2*log(2*NL/d)/d; %%% d reliable observation for adaptive
end

if valtype >= 0
    tau = 2*log(2*N/minobs)/d;
end
%%%%% end of preparations
	

%%%% algorithm
repcand = repinit;

dnew = zeros(d,K);
freq = zeros(1,K);

repcandnew = zeros(d,L);
repfreq = zeros(1,L);

avenres = 0;
avenapp = 0;
Sguess = zeros(1,2);
avsc2=zeros(1,K);

%%% inner products etc. fast
ip=dico'*data;
absip=abs(ip);
signip=sign(ip);
[~,I] = sort(absip,1,'descend');
gram=dico'*dico;

%%% loop over all training signals
for n=1:N

    %%% basic ITKrM steps
    
    lastwarn('');
    coeff = gram(I(1:S,n),I(1:S,n))\ip(I(1:S,n),n);
    %%% if the line above creates a warning (instable)
    %%% recalculate stably but more expensively via pinv
    [~, msgid] = lastwarn;
    if ~isempty(msgid)
        if strcmp(msgid, 'MATLAB:nearlySingularMatrix') || ...
            strcmp(msgid, 'MATLAB:singularMatrix')
            coeff = pinv(gram(I(1:S,n),I(1:S,n)))*ip(I(1:S,n),n);
        end
    end
    
    app = dico(:,I(1:S,n))*coeff;
    res = data(:,n)-app;
    dnew(:,I(1:S,n))=dnew(:,I(1:S,n))+ real(res*signip(I(1:S,n),n)');
    dnew(:,I(1:S,n))=dnew(:,I(1:S,n)) + dico(:,I(1:S,n))*diag(absip(I(1:S,n),n));
    %%% end basic ITKrM

    %%% counter for adaptivity
    enres = res'*res;
    enapp = app'*app;

    if valtype == -1
        %%%% simple counting
        freq(I(1:S,n))=freq(I(1:S,n))+ones(1,S);
    elseif valtype == 0
        %%%% counting above threshold
        freq(I(1:S,n))=freq(I(1:S,n))+ones(1,S).*(abs(coeff').^2>enres*tau + enapp/d);
    else
        %%%% weighted counting above threshold
        freq(I(1:S,n))=freq(I(1:S,n))+abs(coeff').^2.*(abs(coeff').^2>enres*tau + enapp/d);      
    end
    %%% end counter

    %%%% steps for candidates
    repip=repcand'*res;
    [maxrepip,ind]=max(abs(repip));
    repcandnew(:,ind) = repcandnew(:,ind) + res*sign(repip(ind));
    
    if maxrepip^2 > enres*candtau        %%% candidate counter
        repfreq(ind) = repfreq(ind) + 1;    
    end

    %%% after NL signals normalise candidates, restart candidate it.
    if mod(n,NL) == 0  && n < NL*m
        scale = sum(repcandnew.*repcandnew);
        repcand = repcandnew *diag(1./sqrt(scale));
        repcandnew=zeros(d,L);
        if valtype >= 0 
            repfreq = zeros(1,L);
        end
    end

    %%%% steps for estimating the sparsity level
    resip = dico(:,I(S+1:K,n))'*res;
    Sguess(1) = Sguess(1) + length(find(abs([coeff;resip]) > sqrt(enapp/d+enres*2*log(4*K)/d)));
    Sguess(2) = Sguess(2) + length(find(abs(coeff) > sqrt(enapp/d+enres*2*log(4*K)/d)));

    %%% other statistics
    avenres=avenres + enres;
    avenapp = avenapp + enapp;
    avsc2 = avsc2 + sort(abs([coeff;resip]),'descend').^2';

end

avenres = avenres/N;
avenapp = avenapp/N;

scale = sum(dnew.*dnew);
nonzero = find(scale > 0.001*avenapp/d);
iszero = find(scale <= 0.001*avenapp/d);
% unused atoms will be removed outside
freq(iszero) = 0;
dnew(:,nonzero)=dnew(:,nonzero)*diag(1./sqrt(scale(nonzero)));
dico(:,nonzero)=dnew(:,nonzero); 

scale = sum(repcandnew.*repcandnew);
repcand = repcandnew * diag(1./sqrt(scale));

if valtype >= 0
    repfreq = repfreq/d;
end

if valtype == 2
    freq = freq/(minobs*avenapp)*S;
elseif valtype >= 0
    freq = freq/minobs;
end

avsc2=avsc2/N;
Sguess = Sguess/N;

