function [dico,freq,npruned]=prune_coherent(dico,freq,mode,mumax)

% syntax: [pdico,pfreq,nprune]=prune_coherent(dico,freq,mode,mumax)
%
% prune coherent atoms in a dictionary
%
% input:
% dico... d x K dictionary matrix with coherent atoms to replace
% freq... m x K vector of usage scores for each dictionary atom
%               freq(1,:) should contain most recent scores
% mode... combination mode 'a' merge atoms in 1:1 ratio (default),
%                          'm' merge atoms according to usage score, 
%                          'd' delete less often used one.                           
% mumax... maximally allowed coherence between atoms - default 1./sqrt(2)

% output:
% pdico... d x Kn pruned dictionary 
% freq...  m x Kn vector of usage scores of dictionary atoms
% npruned... number of pruned atoms
%
% Karin Schnass 28.05.18 


%%%% preparations and basic safeguarding
if(nargin < 1)
    disp('[pdico,pfreq,nprune]=prune_coherent(dico,freq,mode,mumax)');
    dico=[];
    return;
end

if nargin < 2
    freq = ones(1,K);
end

[~,K]=size(dico);
[~,Kf]=size(freq);
if Kf ~= K
    disp('matrix sizes do not match')
    npruned = 0;
    return;
end

if nargin < 3;
    mode = 'a';
end

if ~ismember(mode,['c','d','a','m'])
    disp('unknown mode, default used');
    mode = 'a';
end

if nargin < 4;
    mumax=1/sqrt(2);
end

if K == 1 || mumax >=1
    npruned = 0;
    return;
end
%%%% end of preparations 

%%%% pruning
prune=[];
holgram = abs(dico'*dico-eye(K));
maxcorr=max(max(holgram));
while maxcorr>mumax 
    [rowi,coli]=find(holgram==maxcorr);
    rowi=rowi(1);
    coli=coli(1);
    h=sign(dico(:,rowi)'*dico(:,coli));
    if ismember(mode,['c','d'])
        %%% compare most recent score
        if freq(1,rowi)>freq(1,coli)  
            newatom = dico(:,rowi);
        else
            newatom = dico(:,coli);
        end
    elseif mode == 'm'
        %%% correctly signed sum weighted according to most recent score
        newatom = freq(1,rowi)*dico(:,rowi)+ h*freq(1,coli)*dico(:,coli);
    else 
        %%% correctly signed sum 
        newatom = dico(:,rowi)+ h*dico(:,coli);
    end
    dico(:,rowi)= newatom;
    if norm(newatom)>0
        dico(:,rowi)= newatom/norm(newatom);
    end
    freq(1,rowi) = freq(1,rowi)+freq(1,coli);
    prune = [prune, coli];
    holgram(:, [rowi,coli]) = 0;
    holgram([rowi,coli],:) = 0;
    maxcorr=max(max(holgram));
end
dico(:,prune)=[];
freq(:,prune)=[];
npruned = length(prune);


