function [dico,freq,nrep,repcand,repfreq]=replace_coherent(dico,repcand,freq,repfreq,mumax,mode)

% syntax: [dico,freq,repcand,repfreq,nrep]=replace_atoms(dico,freq,repcand,repfreq,mode,mumax)
%
% replace coherent atoms in a dictionary
%
% input:
% dico... d x K dictionary matrix with coherent atoms to replace
% repcand... d x L matrix of replacement candidates 
% freq... m x K vector of usage scores for each dictionary atom
%                   freq(1,:) should contain most recent scores
% repfreq... 1 x L vector of usage score of replacement candidates
%
% mumax... maximally allowed coherence between atoms - default 1./sqrt(2)
% mode... combination mode 'a' merge atoms in 1:1 ratio (default),
%                          'm' merge atoms according to usage score, 
%                          'd' delete less often used one. 
%                          
% output:
% dico... d x K incoherent dictionary with replaced atoms
% freq... m x K vector of usage scores of old and new atoms
% nrep... number of replaced atoms
% repcand... d x l matrix of left over replacement candidates                   
% repfreq... 1 x l vector of usage score of left over candidates
%
% Karin Schnass 28.05.18

%%%% preparations and basic safeguarding
if(nargin < 2)
    disp('[dico,freq,repcand,repfreq,nrep]=replace_coherent(dico,repcand,freq,repfreq,mumax,mode)');
    return;
end

[d,K]=size(dico);
[dl,L]=size(repcand);

if nargin < 3
    freq = ones(1,K);
end

if nargin < 4
    repfreq = ones(1,L);
end

if nargin < 5
    mumax = 1/sqrt(2);
end

if nargin < 6
    mode ='a';
end
    
[m, Kf] = size(freq);
[mr,Lf] = size(repfreq);

if Kf ~= K || dl~=d || Lf~=L || mr>1
    disp('check size of input matrices, >> help replace_coherent');
    nrep = 0;
    return;
end

if ~ismember(mode,['c','d','a','m'])
    disp('unknown mode, default used');
    mode = 'a';
end
%%% end of preparations


nrep = 0;

%%% sort candidates by usefulness score
[repfreq,repind] = sort(repfreq,'descend');
repcand = repcand(:,repind);

%%% calculate hollow gram matrix 
holgram = abs(dico'*dico - eye(K));
maxcorr = max(max(holgram));
while maxcorr > mumax && ~isempty(repfreq)
    [rowi,coli] = find(holgram==maxcorr);
    rowi = rowi(1);
    coli = coli(1);
    repatoms = dico(:,[rowi,coli]);
    h=sign(dico(:,rowi)'*dico(:,coli));
    
    if ismember(mode,['c','d'])
        %%% compare most recent score
        if freq(1,rowi)>freq(1,coli)  
            newatom = dico(:,rowi);
        else
            newatom = dico(:,coli);
        end
    elseif mode == 'm'
        %%% correctly signed sum weighted according to most recent score
        newatom = freq(1,rowi)*dico(:,rowi)+ h*freq(1,coli)*dico(:,coli);
    else 
        %%% correctly signed sum 
        newatom = dico(:,rowi)+ h*dico(:,coli);
    end
    dico(:,[rowi,coli]) = [newatom,newatom];
    if norm(newatom)>0
        dico(:,[rowi,coli]) = [newatom,newatom]/norm(newatom);
    end
    rep_found = 0; 
    %%% find incoherent replacement candidate
    while ~rep_found && ~isempty(repfreq) 
        coh_repcand = max(abs(dico'*repcand(:,1)));
        %%% check coherence of candidate
        if coh_repcand < maxcorr
            %%% if candidate is less coherent than orignal, 
            %%% update dictionary and number of replaced atoms
            freq(:,rowi) = freq(:,rowi)+freq(:,coli);
            dico(:,coli) = repcand(:,1);  
            if coh_repcand < mumax
                freq(:,coli) = repfreq(1)*ones(m,1);
            else
                freq(:,coli) = zeros(m,1);
            end
            rep_found = 1;
            nrep = nrep + 1;
        end
        %%% discard candidate
        repcand(:,1)=[];
        repfreq(1)=[];
    end
    %%% if no less coherent replacement candidate is found, 
    %%% keep old atoms
    if ~rep_found
        dico(:,[rowi,coli]) = repatoms;
    end
    %%% update hollow gram matrix and maxcorr
    holgram = abs(dico'*dico - eye(K));
    maxcorr=max(max(holgram));
end



