
%%%%%%%%%% settings %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%% script parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rng(42);        %%% random seed
display = 1;    %%% display messages, jumping points, etc.
runs = 10;      %%% number of trials
savefile = 'mysavefile.mat';


%%%%% signal parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
picname = 'mandrill.png';   %%% picture to extract patches from
sigma = 0;                  %%% noiselevel 
s1 = 8;                     %%% patchheight
s2 = s1;                    %%% patchwidth
d = s1*s2;   
normsig = 0;                %%% normalise patches (may blow up noise) or not 
remove_foreground = 1;      %%% remove contribution average of each patch
         

%%%%% algorithm parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
N = 70000;          %%% signals per iteration
maxit = 100;        %%% number of iterations
inittype = 'r';     %%% 'r' random initialisation (default), 'd' data
Ke = d;             %%% initial estimate of dico size (atom number)
Se = 1;             %%% initial estimate of sparsity level
                            %%% should satisfy Se <= Ke/log(Ke)        

%%%%% adaptivity parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
mumax = 0.7;                %%% maximally allowed coherence 
combimode = 'm';            %%% 'm' merge, 'a' add, 'd' delete coherent atoms
minobs = round(d*log(d));   %%% prune atoms less often used than minobs
m = round(log(d));          %%% protected runs for newly added atoms
L = round(log(d));          %%% number of candidates learned per iteration

%%% adaptivity limits - budget constraints %%%%%%%%%%%%%%%%%%%%%%%%
Smax = round(d/log(d));      %%% maximally allowed sparsity level 
Kmax = round(d*log(d));      %%% maximally allowed number of atoms

Ke = min([Kmax,Ke]);
Se = min([Se, Smax,Ke]);

%%% end of settings %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%% load picture 
pic = imread(strcat('images/',picname));
pic = im2double(pic);
[d1,~,d3]=size(pic);
if d3 > 1
    pic = 0.2989*pic(:,:,1)+0.5870*pic(:,:,2)+0.1140*pic(:,:,3);
end
pic = imresize(pic,256/d1);   %%% get 256x256 or so 
[d1,d2] = size(pic);
Nmax = (d1-s1+1)*(d2-s2+1);
N = min(Nmax,N);
%%%% end of loading picture

%%% create noise patterns for each run 
noise = randn(d1,d2,runs);

%%%% initialisation of dictionary and replacement candidates
dinit = zeros(runs,d,Ke);
repcandinit = zeros(runs,d,maxit*L);

if Nmax < Ke+maxit*L
    inittype = 'r';
end

if inittype == 'd' 
    for run=1:runs
        %%% create data from (noisy) image
        picn = pic + sigma/255*noise(:,:,run);
        [Y, ~] = pic2patches(picn,s1,s2);
        %%% pick random signals
        p = randperm(Nmax);
        dinit(run,:,:)=Y(:,p(1:Ke));
        repcandinit(run,:,:) = Y(:,p(Ke+1:Ke+maxit*L));
    end
else
    dinit = randn(runs,d,Ke);
    repcandinit = randn(runs,d,maxit*L);
end
%%%% end of initialisation of dictionary and replacement candidates

%%%% dictionary learning statistics to track
rdicos = zeros(d,Kmax,runs);
npruned_coh = zeros(runs,maxit);
npruned_uu = zeros(runs,maxit);
nadded = zeros(runs,maxit);
Khistory = Ke*ones(runs,maxit);

Shistory = Se*ones(runs,maxit);
%%% end of statistics to track

%%% approximation statistics for clean patches of learned dico
Sapp = 12;          %%% maximal sparsity level for approximation
meanerr = zeros(runs,Sapp);
meanerr2 = zeros(runs,Sapp);
rel_meanerr = zeros(runs,Sapp);
rel_meanerr2 = zeros(runs,Sapp);

th1 = 0.1; 
th2 = sqrt(2)*0.1;
th3 = 0.2;

Sfor_relerr1 = zeros(runs,Sapp);
Sfor_relerr2 = zeros(runs,Sapp);
Sfor_relerr3 = zeros(runs,Sapp);
%%% end of approximation statistics 

%%% start of runs
for run=1:runs
    
    %%% create data from (noisy) image
    picn = pic + sigma/255*noise(:,:,run);
    [Y, ~] = pic2patches(picn,s1,s2);
    if sigma > 0 && display
        figure,
        imagesc(picn);
    end
    %%% end of data creation

    %%% preprocess patches => training data Y
    if normsig
        scale = sqrt(sum(Y.*Y,1));        % normalise patches
        Y=Y.*(ones(d,1)*(1./scale));    % (may blow up noise)                                 % or don't
    end
    if remove_foreground
        Y=Y-(1/sqrt(d))*ones(d,1)*sum(Y,1)/sqrt(d);  % remove foreground
    end
    ensig = mean(sum(Y.*Y,1));
    %%% end of preprocessing 
    
    %%% initialise dictionary size, sparsity level and usage score
    rdico = zeros(d,Ke);
    rdico(:,:) = dinit(run,:,:);
    rdico = rdico*diag(1./sqrt(sum(rdico.*rdico)));
    Kr=Ke;
    Sr=Se;
    rfreqm=ones(m,Kr);
    %%% end of initialisation
   
    %%% start of iterations for adaptive ITKrM 
    for it=1:maxit;
        
        %%% initialise replacement candidates
        repcand = zeros(d,L);
        repcand(:,:) = repcandinit(run,:,(it-1)*L+1:it*L);
        repcand=repcand*diag(1./sqrt(sum(repcand.*repcand)));
        
        %%%% one iteration of itkrm
         p=randperm(Nmax);  %%% change order of signals
        [rdico,rfreq,repcand,repfreq,Sguess,avenres,avsc2] = itkrmplus_1it(rdico,Y(:,p(1:N)),Sr,repcand,0,minobs);
        
        %%% update atom scores
        rfreqm(2:m,:)=rfreqm(1:m-1,:);
        rfreqm(1,:)=rfreq;
        
        if display
            SNRest = (ensig-avenres)/avenres
            Kest = length(find(rfreqm(1,:)>1))
            rstat = [max(rfreqm(1,:)), median(rfreqm(1,:)),min(rfreqm(1,:))]
            sort(repfreq,'descend')
        end
        
        %%%%%% prune coherent and unused atoms %%%%%%%%%%%%
        
        %%% prune coherent atoms
        if it > m && Sr>1
            [rdico,rfreqm,npruned]=prune_coherent(rdico,rfreqm,combimode,mumax);
            npruned_coh(run,it) = npruned;
        else
            %%% possibility to allow higher coherence 
            %%% if Sr = 1 || it < memory
            [rdico,rfreqm,npruned]=prune_coherent(rdico,rfreqm,combimode,mumax);
            npruned_coh(run,it) = npruned;
        end

        %%% prune unused atoms
        %%% maximal score in last m runs < 1
        if it >= 2*m
            %%% best score in last iteration 
            maxscore = max(rfreqm(1,:));
            if maxscore > 1
                [rdico,rfreqm,npruned]=prune_unused(rdico,rfreqm,min(1,maxscore/2));
                npruned_uu(run,it) = npruned;
            else 
                %%%%% safeguard for ill-chosen number of observations
                %%%%% (compared to number of training signals)
                %%%%% and data that is not sparse
                [rdico,rfreqm,npruned]=prune_unused(rdico,rfreqm,maxscore/sqrt(d));
                npruned_uu(run,it) = npruned;
                if display
                    disp('warning: not enough training signals or data not sparse')
                end
            end
        end
        %%%%%% end of pruning %%%%%%%%%%%%%%%
        
        %%%%%% add promising candidates 
        if it>= m && it < maxit-3*m 

            [rdico,rfreqm,nadd]=add_atoms(rdico,rfreqm,repcand,repfreq,mumax,1);
            nadded(run,it) = nadd;

        end
        %%% end of adding
        
        Kr=size(rdico,2);
        Khistory(run,it) = Kr;
        
        %%%% enforce a maximal size if required
        if Kr > Kmax
            rdico = rdico(:,1:Kmax);
            Kr = Kmax;
        end
        %%% end of size check
            
        %%% adapt sparsity level
        if it > m
            Sr = Sr + sign((round(Sguess(1)))-Sr);
            Sr = min([Smax,Sr]);
            Sr = max(Sr,1);
            
            Shistory(run,it)=Sr;
        else 
            Sr=min(Sr,Kr);
        end
        %%% end of adapting sparsity level
       
        %%%% show progress
        if display
            
            subplot(411)
            imagesc(showdico(rdico,s1,2,round(sqrt(Kr/4)),1,-0.5));
            caxis([-0.5,0.5])
            title(num2str([it, Sr, Kr,nadded(run,it),npruned_coh(run,it),npruned_uu(run,it)]));
            
            subplot(412), 
            plot(1:Kr,rfreqm(1,:),'rx',1:Kr,max(rfreqm),'mx');
            axis([0,Kr,0,5])
            xlabel('r: last score, m: max score,')
            
            subplot(413), 
            rmu=(max(abs(rdico'*rdico-eye(Kr))));
            plot(1:Kr,rmu,'mx');
            xlabel('coherence')
                
            subplot(414), 
            supp = min(round(d/sqrt(2*log(2*(Kr+1)))),length(avsc2));
            plot(1:supp,avsc2(1:supp)/ensig,'rx')
            xlabel('r: sorted av. coeff. and inner prod. squared/av. energy')
            drawnow;
        end
        %%% end of progress display
        
        %%% save runs
        if exist('savefile','var')
            save(savefile);
        end 
    end
    
    rdicos(:,1:Kr,run) = rdico;
    
    %%%% use dictionary for sparse approximation of clean patches
    [patches, ~] = pic2patches(pic,s1,s2);
    
    %%% absolute errors
    apperr = OMP_Shist([(1/sqrt(d))*ones(d,1),rdico],patches,Sapp);
    meanerr(run,:) = mean(apperr,2);    
    %%%% for Fig.7 plot(mean(meanerr2/mean(scale.^2))
    meanerr2(run,:) = mean(apperr.^2,2);
    
    %%% relative errors
    scale=sqrt(sum(patches.*patches,1)); 
    rel_apperr = apperr.*(ones(Sapp,1)*(1./scale));
    rel_meanerr(run,:) = mean(rel_apperr,2);
    rel_meanerr2(run,:) = mean(rel_apperr.^2,2);
    Sfor_relerr1(run,:) = sum(ceil(th1-rel_apperr)')';
    Sfor_relerr2(run,:) = sum(ceil(th2-rel_apperr)')';
    Sfor_relerr3(run,:) = sum(ceil(th3-rel_apperr)')';

end

%%% save all compactly
if exist('savefile','var')
    clear repcandinit; clear dinit; clear Y; clear patches; 
    save(savefile);
end 

