
%%%%%%%%%% settings %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%% script parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rng(42);        %%% random seed
display = 1;    %%% display messages, jumping points, etc.
runs = 20;      %%% number of trials
savefile = 'mysavefile.mat';

%%%% signal parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
dicotype = 'rd';    %%% 'rd' random dictionary, 'dd' dirac, dct (K<=2d)
d = 128;            %%% signal dimension
K = 1.5*d;          %%% dictionary size
S = 6;              %%% sparsity level (S-Sp):S:S+Sp = 1:2:1
Sp = 0;    
b = 0.1;            %%% max. decay of sparse coefficients c_i=(1-b)^i
rho = 0.25/sqrt(d); %%% noiselevel
outliers = 0.05;    %%% percentage of outliers (= pure Gaussian noise)
normsig = 0;        %%% normalise training signals or not

%%%%% algorithm parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
N = 120000;         %%% signals per iteration
maxit = 100;        %%% number of iterations
inittype = 'r';     %%% 'r' random initialisation (default), 'd' data
Ke = d;             %%% initial estimate of dico size (atom number)
Se = 1;             %%% initial estimate of sparsity level

%%%%% adaptivity parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
mumax = 0.7;                %%% maximally allowed coherence 
combimode = 'm';            %%% 'm' merge, 'a' add, 'd' delete coherent atoms
minobs = round(d*log(d));   %%% prune atoms less often used than minobs
m = round(log(d));          %%% protected runs for newly added atoms
L = round(log(d));          %%% number of candidates learned per iteration

%%% adaptivity limits - budget constraints %%%%%%%%%%%%%%%%%%%%%%%%%
Smax = round(d/log(d));      %%% maximally allowed sparsity level 
Kmax = round(d*log(d));      %%% maximally allowed number of atoms

Ke = min([Kmax,Ke]);
Se = min([Se, Smax,Ke]);
%%% end of settings %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%% create generating dictionary
if dicotype == 'dd'
    K = min(K,2*d);
    dico=[eye(d),idct(eye(d))];
    p=randperm(2*d);
    dico=dico(:,p(1:K));
else
    dico=randn(d,K);
    dico=dico*diag(1./sqrt(sum(dico.*dico)));
end
%%% end of dico generation

%%%% initialisation of dictionary and replacement candidates
dinit = zeros(runs,d,Ke);
repcandinit = zeros(runs,d,maxit*L);

if inittype == 'd'
    Ninit=K*runs+L*maxit*runs; 
    Y=zeros(d,Ninit);
    Sl=max(1,S-Sp);
    Ninito4 = ceil(Ninit/4);
    Y(:,1:Ninito4) = makesparsesig(dico,Ninito4,Sl,Sl,b,rho,0);
    Y(:,Ninito4+1:2*Ninito4) = makesparsesig(dico,Ninito4,S+Sp,S+Sp,b,rho,0);
    Y(:,2*Ninito4+1:Ninit) = makesparsesig(dico,Ninit-2*Ninito4,S,S,b,rho,0);
    p = randperm(Ninit);
    Y=Y(:,p);
    if outliers > 0;
        p=randperm(Ninit);
        Nout=round(Ninit*outliers);
        Yout=randn(d,Nout)/sqrt(d);
        Y(:,p(1:Nout))=Yout;
    end
    for run=1:runs
        dinit(run,:,:)=Y(:,1:Ke);
        Y(:,1:Ke) = [];
        repcandinit(run,:,:) = Y(:,1:L*maxit);
        Y(:,1:L*maxit)=[];
    end
else
    dinit = randn(runs,d,Ke);
    repcandinit = randn(runs,d,maxit*L);
end
%%%% end of initialisation of dictionary and replacement candidates

%%%% statistics to track
bestipr = zeros(runs,maxit,K);
natfound = zeros(runs,maxit);

npruned_coh = zeros(runs,maxit);
npruned_uu = zeros(runs,maxit);
nadded = zeros(runs,maxit);
Khistory = Ke*ones(runs,maxit);

Shistory = Se*ones(runs,maxit);
%%% end of statistics to track

%%% start of runs
for run=1:runs
    
    %%% initialise dictionary (size), sparsity level and usage score
    rdico = zeros(d,Ke);
    rdico(:,:) = dinit(run,:,:);
    rdico = rdico*diag(1./sqrt(sum(rdico.*rdico)));
    Kr=Ke;
    Sr=Se;
    rfreqm=ones(m,Kr);
    %%% end of initialisation
   
    %%% start of iterations for adaptive ITKrM 
    for it=1:maxit
        
        %%% create training data Y
        Sl=max(1,S-Sp);
        Y = zeros(d,N);
        No4 = ceil(N/4);
        Y(:,1:No4) = makesparsesig(dico,No4,Sl,Sl,b,rho,0);
        Y(:,No4+1:2*No4) = makesparsesig(dico,No4,S+Sp,S+Sp,b,rho,0);
        Y(:,2*No4+1:N) = makesparsesig(dico,N-2*No4,S,S,b,rho,0);
        p=randperm(N);
        Y=Y(:,p);
        if outliers > 0;
            p=randperm(N);
            Nout=round(N*outliers);
            Yout=randn(d,Nout)/sqrt(d);
            Y(:,p(1:Nout))=Yout;
        end
        %%% end of create data
        
        %%% preprocess training data Y
        if normsig
            scale = sqrt(sum(Y.*Y,1));        % normalise data
            Y = Y.*(ones(d,1)*(1./scale));    % (may blow up noise)
        end
        ensig = sum(sum(Y.*Y))/N;
        
        %%% initialise replacement candidates
        repcand = zeros(d, L);
        repcand(:,:) = repcandinit(run,:,(it-1)*L+1:it*L);
        repcand=repcand*diag(1./sqrt(sum(repcand.*repcand)));
        
        %%%% one iteration of itkrm %%% 
        [rdico,rfreq,repcand,repfreq,Sguess,avenres,avsc2] = itkrmplus_1it(rdico,Y,Sr,repcand,0,minobs);
        
        %%% update atom scores
        rfreqm(2:m,:)=rfreqm(1:m-1,:);
        rfreqm(1,:)=rfreq;
        
        if display
            SNRest = (ensig-avenres)/avenres
            Kest = length(find(rfreqm(1,:)>1))
            rstat = [max(rfreqm(1,:)), median(rfreqm(1,:)),min(rfreqm(1,:))]
            sort(repfreq,'descend')
        end
        
        %%%%%% prune coherent and unused atoms %%%%%%%%%%%%
        
        %%% prune coherent atoms
        if it > m && Sr>1
            [rdico,rfreqm,npruned]=prune_coherent(rdico,rfreqm,combimode,mumax);
            npruned_coh(run,it) = npruned;
        else
            %%% possibility to allow higher coherence 
            %%% if Sr = 1 || it < memory
            [rdico,rfreqm,npruned]=prune_coherent(rdico,rfreqm,combimode,mumax);
            npruned_coh(run,it) = npruned;
        end

        %%% prune unused atoms
        %%% maximal score in last m runs < 1
        if it >= 2*m
            %%% best score in last iteration 
            maxscore = max(rfreqm(1,:));
            if maxscore > 1
                [rdico,rfreqm,npruned]=prune_unused(rdico,rfreqm,min(1,maxscore/2));
                npruned_uu(run,it) = npruned;
            else 
                %%%%% safeguard for ill-chosen number of observations
                %%%%% (compared to number of training signals)
                %%%%% and data that is not sparse
                [rdico,rfreqm,npruned]=prune_unused(rdico,rfreqm,maxscore/sqrt(d));
                npruned_uu(run,it) = npruned;
                if display
                    disp('warning: not enough training signals or data not sparse')
                end
            end
        end
        %%%%%% end of pruning %%%%%%%%%%%%%%%
          

        %%%%%% add promising candidates 
        if it>= m && it < maxit-3*m 

            [rdico,rfreqm,nadd]=add_atoms(rdico,rfreqm,repcand,repfreq,mumax,1);
            nadded(run,it) = nadd;

        end
        %%% end of adding  
            
        Kr=size(rdico,2);
        Khistory(run,it) = Kr;
        
        %%%% enforce a maximal size if required
        if Kr > Kmax
            rdico = rdico(:,1:Kmax);
            Kr = Kmax;
        end
        %%% end of size check
        
            
        %%% adapt sparsity level
        if it > m
            Sr = Sr + sign((round(Sguess(1)))-Sr);
            Sr = min([Smax,Sr]);
            Sr = max(Sr,1);
            
            Shistory(run,it)=Sr;
        else 
            Sr=min(Sr,Kr);
        end
        %%% end of adapting sparsity level
       
        bestipr(run,it,:) = max(abs(rdico'*dico));
        natfound(run,it) = length(find(bestipr(run,it,:)>0.99));
        rmu=(max(abs(rdico'*rdico-eye(Kr))));
        rmumax=max(rmu);
        
        %%%% show progress
        if display
            subplot(411),
            plot(max(abs(rdico'*dico)),'rx');
            title(num2str([it, Sr, Kr,nadded(run,it),npruned_coh(run,it),npruned_uu(run,it)]));
            xlabel(['atoms found: ', num2str(length(find(bestipr(run,it,:)>0.99)))]);

            subplot(412), 
            plot(1:Kr,rfreqm(1,:),'rx',1:Kr,max(rfreqm),'mx');
            xlabel('r: last score, m: max score,')
            
            subplot(413), 
            plot(1:Kr,rmu,'mx');
            xlabel('coherence')
                
            subplot(414), 
            supp = min(round(d/sqrt(2*log(2*(K+1)))),length(avsc2));
            plot(1:supp,avsc2(1:supp)/ensig,'rx')
            xlabel('r: sorted av. coeff. and inner prod. squared/av. energy')
            drawnow;
        end
        %%% end of progress display
        
         %%% save results up to current run and iteration
        if exist('savefile','var')
            Y = Y(:,1:1000);  
            save(savefile);
        end 
    end
    %%% end of adaptive itkrm iteration 
end
%%% end of run

%%% save all compactly
if exist('savefile','var')
    clear repcandinit; clear dinit;
    save(savefile);
end 

