
%%%%% settings %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%% script parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
rng(42);        %%% random seed
display = 1;    %%% display messages, jumping points, etc.
runs = 20;      %%% number of trials
savefile = 'mysavefile.mat';

%%%% signal parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
dicotype = 'rd';    %%% 'rd' random dictionary, 'dd' dirac, dct (K<=2d)
d = 128;            %%% signal dimension
K = 1.5*d;          %%% dictionary size
S = 6;              %%% sparsity level (S-Sp):S:S+Sp = 1:2:1
Sp = 0;    
b = 0.1;            %%% max. decay of sparse coefficients c_i=(1-b)^i
rho = 0.25/sqrt(d); %%% noiselevel
outliers = 0.05;    %%% percentage of outliers (= pure Gaussian noise)
normsig = 0;        %%% normalise training signals or not

%%%%% algorithm parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
N = 120000;         %%% signals per iteration
maxit = 100;        %%% number of iterations
inittype = 'r';     %%% 'r' random initialisation (default), 'd' data
Se = S;             %%% estimated sparsity level used in algorithm
           
%%%%% replacement parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
replace = 'c';      %%% n... no replacement 
                    %%% r... replace with random atoms
                    %%% c... replace with learned candidates 
mumax = 0.7;        %%% maximally allowed coherence 
combimode = 'm';    %%% 'm' merge, 'a' add, 'd' delete
L = round(log(d));  %%% number of candidates learned per iteration
minobs = 1;         %%% replace atoms less often used than minobs

%%% end of settings %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%% create generating dictionary
if dicotype == 'dd'
    K = min(K,2*d);
    dico=[eye(d),idct(eye(d))];
    p=randperm(2*d);
    dico=dico(:,p(1:K));
else
    dico=randn(d,K);
    dico=dico*diag(1./sqrt(sum(dico.*dico)));
end
%%% end of dico generation

%%%% initialisation of dictionary estimate and replacement candidates
dinit = zeros(runs,d,K);
repcandinit = zeros(runs,d,maxit*L);

if inittype == 'd'
    Ninit=K*runs+L*maxit*runs; 
    Y = zeros(d,Ninit);
    Sl=max(1,S-Sp);
    Ninito4 = ceil(Ninit/4);
    Y(:,1:Ninito4) = makesparsesig(dico,Ninito4,Sl,Sl,b,rho,0);
    Y(:,Ninito4+1:2*Ninito4) = makesparsesig(dico,Ninito4,S+Sp,S+Sp,b,rho,0);
    Y(:,2*Ninito4+1:Ninit) = makesparsesig(dico,Ninit-2*Ninito4,S,S,b,rho,0);
    p = randperm(Ninit);
    Y=Y(:,p);
    if outliers > 0;
        p=randperm(Ninit);
        Nout=round(Ninit*outliers);
        Yout=randn(d,Nout)/sqrt(d);
        Y(:,p(1:Nout))=Yout;
    end
    for run=1:runs
        dinit(run,:,:)=Y(:,1:K);
        Y(:,1:K) = [];
        repcandinit(run,:,:) = Y(:,1:L*maxit);
        Y(:,1:L*maxit)=[];
    end
else
    dinit = randn(runs,d,K);
    repcandinit = randn(runs,d,maxit*L);
end
%%%% end of initialisation of dictionary and replacement candidates

%%%% statistics to track
bestipr=zeros(runs,maxit,K);
natfound = zeros(runs,maxit);
if ismember(replace,['c','r']) 
    nrep_coh = zeros(runs,maxit);
    nrep_uu = zeros(runs,maxit);
end
%%% end of statistics to track

%%% start of runs
for run=1:runs
    
    %%% initialise dictionary size, sparsity level and usage score
    rdico = zeros(d,K);
    rdico(:,:) = dinit(run,:,:);
    rdico = rdico*diag(1./sqrt(sum(rdico.*rdico)));
    
    %%% start of iterations of ITKrM with(out) replacement
    for it=1:maxit
        
        %%% create training data Y
        Sl=max(1,S-Sp);
        Y = zeros(d,N);
        No4 = ceil(N/4);
        Y(:,1:No4) = makesparsesig(dico,No4,Sl,Sl,b,rho,0);
        Y(:,No4+1:2*No4) = makesparsesig(dico,No4,S+Sp,S+Sp,b,rho,0);
        Y(:,2*No4+1:N) = makesparsesig(dico,N-2*No4,S,S,b,rho,0);
        p=randperm(N);
        Y=Y(:,p);
        if outliers > 0;
            p=randperm(N);
            Nout=round(N*outliers);
            Yout=randn(d,Nout)/sqrt(d);
            Y(:,p(1:Nout))=Yout;
        end  
        %%% end of create data
        
        %%% preprocess training data Y
        if normsig
            scale=sqrt(sum(Y.*Y,1));        %%% normalise data
            Y=Y.*(ones(d,1)*(1./scale));    %%% (may blow up noise)
        end
        ensig = sum(sum(Y.*Y))/N;
        
        %%% initialise replacement candidates
        repcand = zeros(d, L);
        repcand(:,:) = repcandinit(run,:,(it-1)*L+1:it*L);
        repcand=repcand*diag(1./sqrt(sum(repcand.*repcand)));
        
        %%% one iteration of itkrm
        if replace == 'c'
            [rdico,rfreq,repcand,repfreq,~,~,avsc2] = itkrmplus_1it(rdico,Y,Se,repcand,-1);
        else
            [rdico,rfreq,~,~,~,~,avsc2] = itkrmplus_1it(rdico,Y,Se,repcand,-1);
            repfreq = ones(1,L);
        end
      
        if display && replace == 'c'
            [max(rfreq), median(rfreq), min(rfreq)]
            sort(repfreq,'descend')
        end
       
        %%% replacement of coherent and unused atoms
        if ismember(replace, ['r','c']) 
            
            %%% replace coherent atoms
            [rdico,rfreq,nrep,repcand,repfreq] = replace_coherent(rdico,repcand,rfreq,repfreq,mumax,combimode);
            nrep_coh(run,it) = nrep;
            
            %%% replace unused atoms
            if it>=10 && ~isempty(repcand) 
                [rdico,rfreq,nrep,~,~]=replace_unused(rdico,repcand,rfreq,repfreq,minobs);
                nrep_uu(run,it) = nrep;
            end
        end
        %%% end of replacement 
       
        bestipr(run,it,:) = max(abs(rdico'*dico));
        natfound(run,it) = length(find(bestipr(run,it,:)>0.99));
        rmu=(max(abs(rdico'*rdico-eye(K))));
        
        %%%% show progress
        if display
            subplot(411),
            plot(max(abs(rdico'*dico)),'rx');
            if ismember(replace,['r','c'])
                title(num2str([it, nrep_coh(run,it), nrep_uu(run,it)]));
            else
                title(num2str(it));
            end
            xlabel(['atoms found: ', num2str(length(find(bestipr(run,it,:)>0.99)))]);

            subplot(412), 
            plot(1:K,rfreq,'rx');
            xlabel('r: last score')
            
            subplot(413), 
            plot(1:K,rmu,'mx');
            xlabel('coherence');
                
            subplot(414), 
            supp = min(round(d/sqrt(2*log(2*(K+1)))),length(avsc2));
            plot(1:supp,avsc2(1:supp)/ensig,'rx')
            xlabel('r: sorted av. coeff. and inner prod. squared/av. energy')

            drawnow;
        end
        %%% end of progress display
        
        %%% save results up to current run and iteration
        if exist('savefile','var')
            Y = Y(:,1:1000);  
            save(savefile);
        end 
    end
    %%% end of itkrm iteration with(out) replacement
end
%%% end of run

%%% save all compactly
if exist('savefile','var')
    clear repcandinit; clear dinit;
    save(savefile);
end 

